CuLifter: Lifting GPU Binaries to Typed IR
Abstract.
GPU compilers merge all data types into a single unified register file, erasing the type information that binary-analysis tools rely on. We show that type recovery from this untyped register file is the central challenge of GPU binary lifting. We present CuLifter, a SASS-to-LLVM IR lifting framework that recovers register types via constraint propagation with conflict detection, reconstructs explicit control flow, and aggregates multi-instruction patterns. Across eight benchmark suites (24,437 GPU functions in 919 cubins) spanning open-source applications, vendor libraries, and optimized ML runtimes, CuLifter successfully lifts 99.98% of functions to valid LLVM IR. An ablation study confirms that type recovery is the only step required to produce semantically correct IR: disabling it drops the x86 pass rate from 73.8% to 0%, a 73.8 percentage-point drop.
1. Introduction
NVIDIA GPUs dominate high-performance and parallel computing, and CUDA (cuda) has become the de facto standard for GPU programming across ML frameworks (e.g., PyTorch (pytorch), vLLM (vllm)) and HPC components (gpu_survey:2014). Performance-critical CUDA kernels increasingly ship as pre-compiled SASS binaries without source code, including vendor libraries (cuDNN, TensorRT, cuBLAS, NCCL), optimized ML runtimes, and closed-source HPC components.
Binary lifting translates machine code back into a higher-level IR amenable to compiler analysis. For CPU binaries, lifted IR enables a wide range of applications: security analysts use it to detect vulnerabilities in closed-source software (cpu_binary:2012:sp), malware researchers apply taint analysis and symbolic execution on lifted code (cpu_binary:2011:bap), and systems researchers retarget legacy binaries to new architectures (cpu_binary:2000:uqbt) or harden them with control-flow integrity (cpu_binary:2016:asplos). These applications are supported by a mature tool ecosystem: static lifters such as McSema (mcsema) and rev.ng (revng) recover LLVM IR from x86, dynamic lifters such as BinRec (binrec) capture execution traces for recompilation, and a recent systematization of knowledge evaluates these tools across downstream applications (cpu_binary:2020:sp).
No comparable infrastructure exists for GPU binaries. NVIDIA provides disassemblers (nvdisasm, cuObjDump), but these output only flat, untyped assembly without control-flow or type information. NVBit (nvbit) and SASSI (sassi) instrument SASS dynamically and at compile time, respectively, but neither lifts the binary to an analyzable IR. CuAsmRL (cuasmrl:cgo2025) optimizes SASS instruction schedules via reinforcement learning, and CASS (arxiv:25:CASS) transpiles SASS to AMD RDNA3 assembly using LLMs, achieving 37.5% assembly-level translation accuracy. Both operate on SASS but neither produces a typed, retargetable IR suitable for compiler analysis. As a result, the analyses routinely applied to CPU binaries (vulnerability detection, performance modeling, code portability) remain unavailable for the GPU binaries that dominate modern computing workloads.
A key obstacle to GPU binary lifting is type recovery: determining the data type (e.g., Float32, Int32, pointer) that each register holds at each program point. Type recovery is a well-known challenge for CPU binaries as well, where integers and pointers share the same general-purpose registers and memory aliasing obscures data types. A long line of work addresses this problem, progressing from constraint-based inference (tie; secondwrite:pldi2013) to lattice-based propagation (retypd), probabilistic prediction (osprey), and behavior-capturing types (trex:2025:usenix). However, CPU ISAs still provide partial structural cues: separate register files for floating-point and vector values, explicit transfer instructions (movd, cvtsi2sd), and calling conventions that constrain parameter types. GPU ISAs provide far fewer cues. SASS stores all values in a single unified register file, and the same register can be consumed as Float32 by one instruction and as Int32 by the next, with no intervening transfer. Without recovering these implicit type boundaries, a lifter cannot emit IR at all: every LLVM IR instruction requires explicit type annotations, so unresolved types block code generation entirely.
Consider a compiled rsqrt() call:
Register R25 is used as Float32 in lines 1–2 and 6, but reinterpreted as Int32 in line 3, with no transfer instruction between them. A lifter must detect this implicit type boundary and insert an explicit bitcast; otherwise, it cannot assign a single LLVM type to R25.
Our key insight is that GPU binary lifting is a structure-recovery problem. GPU compilation erases three kinds of structure (control flow, instruction-level semantics, and data types), a framing consistent with recent formalizations of binary lifting (fible:2025:oopsla). Among these, type recovery is the most critical and the focus of this paper.
We present CuLifter, a framework that recovers analyzable LLVM IR from NVIDIA SASS. We target LLVM IR because its mature ecosystem of analysis and optimization passes applies directly to the recovered code, and its backends enable retargeting to other architectures (e.g., x86 CPUs). To recover each lost structure, CuLifter applies a targeted technique: it infers register types by propagating type constraints from typed opcodes along def-use edges, reconstructs explicit control flow from SASS’s predicated execution model, and reassembles multi-instruction patterns (e.g., carry chains, reciprocal approximations) into single high-level operations.
We make the following contributions:
-
(1)
We identify type recovery as the central challenge of GPU binary lifting and solve it via constraint propagation with conflict detection.
-
(2)
We adapt -function-based SSA construction (stoutchinin2001psi) to SASS, extending it for SIMT-specific predication and convergence barriers.
-
(3)
We build CuLifter, a complete SASS-to-LLVM IR lifter supporting SM75, SM90, SM100, and SM120, and evaluate it on 24,437 GPU functions across eight benchmark suites. CuLifter successfully lifts 99.98% of functions to valid LLVM IR.
2. Background
2.1. SASS as a Low-Level Representation
| SASS | LLVM IR | |
| operand | physical register | virtual SSA value |
| type system | untyped (unified | strongly typed |
| register file) | (i32, i64, float, ptr, …) | |
| control flow | predicated execution | explicit conditional |
| + unconditional BRA | branch on SSA value | |
| memory access | arch-specific ops | generic load/store |
| (LDG, LDS, LDC, …) | with address space specified | |
| addressing | reg + imm offset | typed pointer |
| tensor op | HMMA, IMMA, | target-specific |
| HGMMA (SM90+) | intrinsics | |
| tensor precision | encoded in opcode | abstracted in |
| suffix (modifiers) | intrinsic signature |
SASS (nvsass) is the native machine code for NVIDIA GPUs. During compilation, CUDA source code is first compiled to PTX (nvptx), a hardware-independent virtual assembly that preserves structured control flow and type information, and then lowered to SASS, a hardware-specific assembly where much of this structure is lost. SASS is optimized for specific GPU architectures (e.g., Turing SM75, Ampere SM80, Hopper SM90), with instruction encoding, register files, memory layout, and synchronization differing across generations. Table 1 summarizes the key structural differences between SASS and LLVM IR.
LLVM IR (llvm) is a strongly typed, platform-independent assembly in static single assignment (SSA) form (ssa). It bridges high-level code and machine-specific assembly, with support for x86, ARM, RISC-V, NVPTX, and others.
2.2. SIMT Execution Model
The NVIDIA GPU architecture is built upon the Single Instruction, Multiple Threads (SIMT) execution model. Threads are grouped into warps of 32 threads that execute in lockstep. When threads within a warp encounter a data-dependent branch, divergence occurs: the warp executes each branch path sequentially, while predicate registers mask off inactive threads. As a result, SASS encodes most conditional behavior through predicated execution rather than explicit branch instructions. Starting with the Volta architecture, NVIDIA introduced Independent Thread Scheduling, which gives each thread its own program counter and call stack, allowing threads to make independent forward progress even after divergence. However, to regain lockstep efficiency once threads reconverge, NVIDIA employs explicit convergence instructions. On Volta and later architectures, BSSY and BSYNC are used; on pre-Volta architectures, SSY and SYNC serve this purpose.
2.3. Predicated SSA and -Functions
Standard SSA construction (ssa) assumes every instruction in a basic block executes unconditionally. Predicated code violates this: two instructions that write the same register under complementary predicates both “execute” in the same block, and standard SSA rename kills the first definition, producing wrong code.
Stoutchinin and de Ferrière (stoutchinin2001psi) introduced the -function (psi-function) for IA-64 predicated code: where a -node selects a value based on which control-flow edge was taken, a -function selects based on a predicate register’s value. A predicated write @P0 R0 = x followed by @!P0 R0 = y becomes , semantically equivalent to a select instruction in LLVM IR. This technique is well-studied in the IA-64 compilation literature (warter1993reverse; carter1999pssa; chuang2003phi).
Both IA-64 and NVIDIA SASS support predicated execution: almost every instruction can be guarded by a predicate register (e.g., p0–p63 in IA-64, P0–P7 in SASS). When the predicate is false, the instruction behaves as a no-op. Both architectures use a unified register file where integer, floating-point, and predicate values are stored; however, the -function concerns only dataflow, not types. To take advantage of the -function, we apply the same principle to the GPU predicated code in Section 5.2.
3. Challenges in GPU Binary Lifting
CPU binary lifters rely on separate register files, explicit branch instructions, and publicly documented ISAs (cpu_binary:2016:asplos; cpu_binary:2011:bap; cpu_binary:2019:survey). SASS provides none of these: its unified register file carries no type information, predication replaces branches for short conditionals, and the ISA is proprietary and changes across GPU generations. Of these lost structures, type recovery is the most critical because it gates IR generation entirely (Section 6.4). We characterize each challenge below, starting with type recovery.
3.1. Type Evidence Is Sparse and Must Be Propagated
Type recovery is the key technical challenge of GPU binary lifting. On CPUs, separate integer and floating-point register files provide type information for free. On SASS, a unified register file holds Int32, Float32, Float64 (as register pairs), pointers, packed Float162, and boolean predicates interchangeably. The register name carries no type information. Because LLVM IR requires explicit types on every instruction, a lifter that cannot distinguish Float32 from Int32 cannot emit valid IR at all.
Most SASS instructions encode type constraints in their opcodes (e.g., FADD operates on floats, ISETP produces a boolean predicate), providing direct type evidence for 90% of all instructions (Section 6.3). However, two challenges remain. First, the remaining type-transparent instructions (moves, bitwise logic, and memory operations) pass values without revealing their type, requiring propagation along def-use chains to resolve them. Second, GPU compilation introduces intentional type conflicts: deliberate reinterpretation of float bits as integers for optimization (lomont2003). We detect over 3.8 million such conflicts across our benchmarks (Section 6.3). Incorrect type assignment produces silent numerical corruption: a 32-bit float and an integer with the same bit pattern differ by orders of magnitude but cause no runtime fault.
As shown in Listing 1, register R25 is defined by FADD (Float32), consumed by MUFU.RSQ (Float32), then read by IADD3 (Int32), and finally used again by FMUL (Float32). A forward-only approach forces Float32 onto IADD3, producing wrong integer arithmetic. The correct solution detects the type conflict () and inserts a bitcast at the boundary.
Key intuition: GPU type recovery is a propagation problem with conflict detection. The density of type seeds and the length of transparent chains determine how far evidence must travel, and empty intersections reveal implicit type boundaries. We formalize this in Section 4.1 and characterize propagation reach empirically in Section 6.3.
3.2. Control Structure Is Not Explicit
Control recovery in SASS is fundamentally challenging due to a structural gap in the representation of control flow. LLVM IR expects an explicit control structure in which conditional execution is implemented strictly via branch instructions. In contrast, SASS uses an implicit model in which almost any instruction can be predicated. Instead of a conditional branch instruction, SASS models conditional branching via predicate-guarded execution of an unconditional BRA. The actual branch conditions are contained implicitly within the predicate register, whose set-predicate instruction could be many instructions away. In addition, this implicit nature of predication further complicates SSA construction, because a predicated definition updates a value only for the threads whose predicate is true, while the remaining threads implicitly preserve the old value, breaking standard data-flow tracking assumptions.
Key intuition: control flow in SASS is heavily carried by predicate flow. To bridge this structural gap and produce a valid SSA form, the lifter must handle the predication semantics and reconstruct the missing explicit logic from predication correctly.
3.3. Semantic Operations Are Collapsed into Multi-Instruction Patterns
GPU compilers collapse high-level operations into multi-instruction patterns whose structure changes across GPU generations. A 64-bit subtraction threads a borrow predicate across two IADD3/IADD3.X instructions; if a lifter handles each instruction in isolation, the carry chain breaks and the wrong result is produced. On SM90, a single cp.async in CUDA becomes an orchestrated sequence: FENCE.VIEW.ASYNC.S issues a memory fence, LDGSTS.E.128 performs a fused global-to-shared copy, and LDGDEPBAR/DEPBAR/BAR.SYNC form an async completion chain. A lifter that treats these independently would decompose LDGSTS into a synchronous load and store and discard the barrier chain as no-ops. SM90 also introduces packed-conversion instructions such as F2FP.PACK_AB_MERGE_C, which converts two Float32 values to Float8, saturates, packs both into the upper 16 bits, and merges with a third operand’s lower 16 bits, all in a single opcode. No direct LLVM intrinsic exists; the lifter must synthesize the full decomposition.
Key intuition: Without pattern recognition, lifted IR is correct in behavior but opaque in intent. We present the full pattern catalog in Section 5.3.
3.4. Implicit Register Dependencies
SASS instructions may access registers that are not explicitly named in the disassembly. In many cases, an operand denotes only the first register of a contiguous register group, while the remaining registers are determined implicitly by the opcode semantics. For example, LDG.E.64 R4, [R2] appears with only two explicit registers, but it actually defines R5:R4 and reads R3:R2. A conventional lifter that models only the explicitly named operands therefore misses half the data dependencies. This directly affects correctness. If another instruction writes R5 between the load and its consumer, the lifter must detect the conflict, but it cannot, because R5 never appears in the disassembly. Getting this wrong silently corrupts values rather than producing an error. Tensor-core instructions demonstrate an extreme case. On SM75, HMMA.16816.F32 names four registers but actually accesses fourteen. On SM90, HGMMA.64x128x16.F32 accesses 164 registers through four named operands.
Key intuition: On CPU ISAs, every accessed register is named in the instruction encoding. On SASS, the lifter must reconstruct hidden registers from per-opcode tables before building data dependencies; otherwise SSA, def-use analysis, and type propagation all operate on incomplete information.
3.5. Device Functions Are Embedded
In CPU binaries, every callable function is a distinct ELF symbol with its own entry point, and lifters discover function boundaries directly from the symbol table. On SASS, the compiler embeds device functions inside the kernel function. They are not inlined (they retain their own control flow with CALL and RET instructions), but they share the same SASS section and have no separate symbol table entry.
After CFG construction, the kernel body and the device function appear as disconnected subgraphs within a single SASS function. CALL does not create a normal CFG edge; it saves the return address to a register pair and branches, but does not duplicate the device function.
Key intuition: Embedded device functions must be recovered as call/return structure, not treated as unreachable CFG fragments; otherwise the lifter loses inter-block definitions that flow back through RET.
4. Solution Approach
This section presents our solutions to the two key technical problems at an algorithmic level, independent of SASS-specific details. Section 5 describes the GPU-specific implementation.
4.1. Type Recovery via Constraint Propagation
We formulate type recovery as a constraint-propagation problem over type sets. Each value carries a set of candidate types , where contains Bool; the integer family plus its unsigned counterparts ; and the floating-point family . Section 5.4 formalizes the lattice and width-based partitioning. All values start unconstrained: .
Phase 1: Detect type seed. Instructions with known type semantics (e.g., floating-point add Float32, integer compare Bool) narrow their operands’ sets. Most SASS instructions are type-seeding (Section 3.1), but the remaining type-transparent instructions create gaps that propagation must bridge.
Phase 2: Propagate type information. Type-transparent instructions (Section 3.1) propagate constraints by intersecting the candidate sets of related operands along def-use edges. A worklist-based fixpoint iterates until no set changes.
Phase 3: Detect conflicts and insert bitcasts. When intersection produces , the algorithm has discovered an implicit type boundary, a point where the compiler reinterprets bits across types with no explicit instruction. The conflict resolution rule is: (1) The value retains its definition-site type. (2) The conflicting use site receives an explicit bitcast. (3) Once marked conflicted, subsequent type assignments overwrite rather than intersect, preventing cascading false conflicts.
For example, Figure 1 traces the recovery phase through Listing 1. A register holding a float value is read as an integer by a subsequent instruction. Seeding establishes the definition-site type as Float32. When the integer use requests Int32, the intersection triggers a conflict: the value keeps Float32, and the integer use receives a bitcast fp32 to i32. A later float use of the same register requires no bitcast because the definition-site type is already correct.
| Instruction | Type set after seed | Emitted IR |
| float add | Float32 | %v = fadd float %v, 1e-8 |
| float rsqrt | Float32 | %u = call float @rsqrt(%v) |
| int subtract | Int32 Float32 | %vi = bitcast float %v to i32 |
| conflict! | %r = sub i32 %vi, 0xd000000 | |
| float multiply | Float32 (no conflict) | %w = fmul float %v, %u |
After convergence, values with multiple remaining candidates are resolved by priority (Int32 preferred), a conservative choice that preserves bit patterns for a subsequent bitcast. The key quantity is propagation reach: how many transparent hops evidence must travel. We measure this across real GPU kernels in Section 6.3.
4.2. Predicated SSA via Select Insertion
We apply the -function technique (Section 2.3) to GPU predicated code: each predicated write is isolated into a synthetic basic block and merged via a select:
This is semantically equivalent to a -function: . On GPUs, three additional SIMT-specific constructs require handling beyond IA-64: (1) predicated thread termination must become a conditional return rather than function exit, (2) dual-predicate branches must separate the branch condition from the convergence annotation, and (3) hardware convergence barriers must be treated as metadata. We detail these in Section 5.2.
5. Recovering Analyzable Structure in CuLifter
This section describes three structure recovery techniques identified in Section 3.
5.1. Recovery Ordering
The two techniques from Section 4, together with pattern normalization, must execute in a specific order as shown in Figure 2: control-flow reconstruction first (to build the CFG with SSA form), semantic operation recovery second (to normalize and to aggregate multi-instruction patterns), and type recovery last (to propagate over the normalized graph). Control-flow reconstruction must come first because both pattern matching and type propagation traverse def-use edges, which require a complete CFG in SSA form. Semantic operation recovery must precede type analysis because the type lattice operates over aggregated values; typing a carry-chain pair before aggregation would mistype the borrow predicate as an independent integer rather than as internal to a 64-bit operation. Stages 1 and 2 are architecture-dependent because instruction encodings, control codes, and predicate semantics differ across GPU generations.
5.2. Control Structure Recovery
5.2.1. SSIR Construction
SSIR (SASS Static IR) is the IR that bridges the architecture-specific and architecture-independent stages. SSIR represents six levels: function metadata, basic blocks, instructions with opcodes and operands, operand type information, CUDA-specific objects (cooperative groups, warps), and synchronization operations. During parsing, SSIR preserves architecture-specific information including nested opcodes (e.g., XMAD.MRG), operand properties (e.g., R0.reuse), and control codes. Instruction normalization expands nested opcodes to 3-address form; one SASS instruction may produce several SSIR instructions.
Several architecture-specific issues are resolved at this stage. Branch targets are realigned from control-code lines to valid instruction addresses. Predicate registers are treated as normal operands, and predicated instructions are isolated into single-instruction basic blocks for downstream SSA construction. Compiler-inserted infinite loops after EXIT are removed; NVCC places these as safety padding to prevent threads from falling through to invalid instructions, but they are not reachable control flow. On Volta and later architectures, instructions with multiple definitions (carry-out predicates) are split into separate instructions.
5.2.2. Implicit Register Expansion
CuLifter expands every instruction to its full set of accessed registers using per-opcode tables. Each register group is treated as a single SSIR operand with sub‑register access. This representation preserves the original type constraints. This expansion is applied in the parsing phase.
For 64-bit operations, the named register is replaced with an explicit register pair: R4 becomes R5:R4. For 128-bit operations (LDG.128, STS.128), a register quad R7:R6:R5:R4 is created. For tensor-core instructions, the expansion is shape-dependent:
-
•
HMMA.16816.F32 (SM75): the shape m16n8k16 determines D=4, A=4, B=2, C=4 registers.
-
•
HGMMA.64x128x16.F32 (SM90): the warp-group shape m64 with N=128 determines D=64, A=4, B=32, C=64 registers.
A register group is treated as a single SSIR operand with sub-register access, so that type analysis can preserve the type constraint (R3:R2 as an int64 address operand in LDG) while assigning types to individual registers (e.g., R3 is later used as int32 by IADD3).
5.2.3. Predicate-Aware CFG Construction
We identify basic-block leaders using standard rules (function entry, branch targets, fall-through successors) plus one GPU-specific rule: whenever the predicate guard changes, a new block starts. This isolates predicated regions for later analysis.
On SASS, conditional control flow is encoded as predicate-guarded execution of unconditional instructions, so the lifter must recover the implicit condition. Unconditional branches become Br terminators. For predicated BRA/EXIT patterns, CuLifter extracts the predicate guard and constructs a CondBr with both taken and fall-through successors. For example, a predicated @P0 EXIT terminates only threads satisfying the condition; the remaining threads fall through to the next instruction. A non-SIMT-aware lifter treats this as unconditional termination, making the fall-through unreachable. SIMT-specific instructions (BSSY, BSYNC, WARPSYNC) are kept as ordinary instructions with metadata for later stages.
A more complex case arises with dual-predicate branches, where two independent predicates control the same instruction:
@!P0 BRA P1 has two predicate roles: @!P0 is the SIMT execution guard, and P1 is the branch condition. We represent this as a nested predicated branch; treating either field as primary produces an incorrect CFG.
5.2.4. Dataflow Graph Construction
On top of the CFG, we build an intra-procedural def-use graph. Nodes are SSIR instructions; edges represent register value flow from definitions to uses. The analysis includes predicate registers and special registers (e.g., SR_TID) so that later passes can reason about SIMT-specific data flow. Because SSA guarantees each register has exactly one definition, the analysis is a simple two-pass scan: the first pass maps each register to its defining instruction; the second links each use to its definition in constant time.
5.2.5. -Functions Construction
The predicated code may introduce multiple definitions of the same register appearing in the same block under mutually exclusive predicates. A naive renaming would keep only the last definition, discarding the others. CuLifter solves this by constructing -functions, adapting the IA-64-predicted SSA technique to GPU SIMT execution.
Let be a basic block. For each architectural register , let be the set of predicated writes to in , where each is a predicate register (or its negation). The combined effect is:
where is the value of coming from preceding blocks (i.e., the -node or live-in value). The semantics: for each thread, the -function selects the whose predicate evaluates to true; if none, it selects .
In practice, CuLifter simplifies the to a binary select chain:
Algorithm 1 shows the -function construction for each basic block after control-flow recovery and before SSA renaming. There are two complications introduced for handling SIMT divergence and convergence:
-
(1)
Per-thread predicates: The -function’s select instruction in LLVM IR naturally models per-lane choice. No additional masking is needed because LLVM’s vector types or scalar-on-SIMT semantics (depending on backend) preserve thread-level distinction.
-
(2)
Convergence barriers: Instructions like BSSY/BSYNC (Volta+) or SSY/SYNC (pre-Volta) mark reconvergence points. CuLifter does not construct -functions that cross these barriers. Instead, it treats each barrier-separated region independently, inserting -nodes at convergence points as needed. This prevents incorrect merging of values from threads that have not yet reconverged.
5.2.6. Device Function Recovery
After CFG construction and before SSA, we recover embedded device functions by identifying disconnected subgraphs unreachable from the entry block. Call targets are resolved through the ELF relocation table. Recovered device functions are either extracted into separate SSIR functions or inlined into the caller.
Inlining is the common case: the CUDA compiler inlines most device functions at the PTX level, so the few that survive to SASS are typically small helpers. Extraction is reserved for functions with many call sites (e.g., math library routines compiled as device functions) to avoid code bloat.
5.3. Semantic Operations Recovery
Stage 2 takes the SSIR from Stage 1 and (1) aggregates instruction sequences implementing a logical operation, and (2) recovers CUDA-level objects from synchronization and communication patterns.
5.3.1. Special Code Pattern Recognition
NVIDIA’s compiler decomposes logical operations into multi-instruction patterns, either for performance or because 64-bit values require 32-bit register pairs. Lifting instruction-by-instruction produces verbose IR that loses the original intent.
CuLifter uses a pattern matcher based on variable unification (shown in Algorithm 2). Each pattern template specifies expected opcodes and symbolic operand names. The matcher scans basic blocks, attempts to map template variables to actual registers (unify), and rewrites matched sequences into aggregated pseudo-instructions. Table 2 lists the code patterns in the matcher.
| Pass / Pattern | Description |
| Code Pattern Normalization | |
| XmadToImad | Rewrites the 3-instruction SM52 XMAD / XMAD.MRG / XMAD.PSL.CBCC idiom into a single IMAD |
| OpModTransform | Expands the .X4 address-scale modifier by inserting an explicit SHL 2 instruction |
| SRSubstitute | Scans constant-memory operands (c[0][offset]) for known SR_* offsets (e.g. SR_TID.X at 0x2C); inserts explicit S2R instructions at first use and rewrites the operand to the fresh temp register |
| ReciprocalNorm | Detects the I2FMUFUIADD/IADD3F2I magic-constant reciprocal chain emitted by NVCC; inserts BITCAST nodes around the integer-add step to preserve correct float/int reinterpretation semantics |
| Cooperative Group Object and Operations Recognition | |
| Group creation | Identified through thread mask manipulation instructions (e.g., WARPSYNC, BAR.SYNC) and predicate configurations that define participation masks. |
| Sync operation | BAR.SYNC with specific barrier IDs indicates thread-block synchronization; WARPSYNC with mask patterns identifies warp-level synchronization. |
| Collective operations | Cooperative reductions manifest as specific shuffle instruction sequences. |
| Type differentiation | Thread-block groups exhibit block-wide BAR.SYNC patterns; warp groups show WARPSYNC 0xFFFFFFFF and shuffle-based communication. |
5.3.2. CUDA Object and Operation Recognition
CUDA-specific objects such as cooperative groups are not represented by explicit SASS opcodes; they emerge through specific synchronization and communication patterns. CuLifter identifies them by scanning for characteristic opcode sequences (shown in Table 2).
5.4. Type Recovery
Stage 3 of CuLifter assigns architecture-independent types to all SSIR values using a constraint-propagation mechanism.
5.4.1. Type Domain and Lattice
As shown in figure 3, we partition the lattice into width-based subsets to enable propagation through type-agnostic instructions. The seeding stage differentiates instructions by both the opcode and the modifier. For example, the 16-bit family contains half and bfloat datatypes, which occupies the same register width but encode different exponent/mantissa layouts. While the opcode of HFMA2 and HFMA2.BF16 all suggests , the modifier disambiguates the operand in HFMA2.BF16 to BF16. Note that some types such as TF32 have no designated type in LLVM IR. In this case, we assign them to the actual type modelled in LLVM IR (e.g., TF32 Int32). This preserves type semantics since any tf32-related instructions accepts Int32 as equivalent data type for TF32.
For each value , we maintain a candidate type set . Initially, is set to a top element representing all types. The lattice is extended with to denote a type conflict.
| Category | Example Signature (Defs Uses) |
| Floating-point ALU | FADD |
| Integer ALU | IADD |
| Compare | ISETP |
| Memory (Global) | LDG |
| Memory (Shared) | LDS |
| Tensor core | HMMA.BF16 |
| Type-transparent | MOV, PACK64, SHR |
5.4.2. Seeding and Propagation Rules
Type evidence originates from two sources.
Seeding instructions.
Opcodes with fixed type semantics (e.g., FADD seeds , ISETP seeds ) narrow the candidate set of their defined value:
where is the type constraint of instruction .
Type-transparent instructions.
Moves, bitwise logic, and memory operations propagate types without revealing them. For a transparent instruction with inputs and output :
Type conversion instructions
(e.g., F2I, I2F) relate source and destination types via a known conversion function :
Table 3 categorizes SASS instructions according to their role.
5.4.3. Data-Flow Propagation
Constraints are propagated along the def-use graph of the SSA-form intermediate representation. For each register , we define a set of possible types initially seeded by its defining instruction. The propagation is a fixed-point iteration:
where returns the required type for given the current types of other registers in constraint . For an equality constraint , we have:
The iteration proceeds in reverse topological order of the def-use graph (from uses to definitions) to accelerate convergence. The process terminates when .
5.4.4. Conflict Detection and Bitcast Insertion
A type conflict occurs when the meet of all propagated constraints for a register becomes :
In this case, no single type satisfies all uses of . The solution is to split the register into two value instances connected by an explicit bitcast. Formally, we replace with two fresh registers and such that:
and insert a conversion between them. The new types are guaranteed non- by construction. In the rare case that a register receives no constraint at all (unreachable code or missing seed), we apply a safe fallback: .
5.4.5. Fixed-Point Algorithm
Algorithm 3 presents the complete fixed-point type propagation. The complexity is where is the number of registers, the number of lattice elements (constant), and the maximum number of propagation hops (bounded by the def-use chain length). In practice convergence requires at most 6 iterations (Section 6.3). Figure 4 gives an example that presents the type states for analyzing the code in listing 1.
5.4.6. Resolving Ambiguous Types
After fixpoint convergence, some values may still have (multiple candidate types). CuLifter applies a fallback prioritizing integer types:
In the rare case, if the constraint set is empty because no type is propagated to it, the system defaults to Int32 to ensure compilable IR generation.
5.5. Lowering to LLVM IR
After type annotation and pattern normalization, we emit LLVM IR through per-opcode handlers. Table 4 shows the mapping: physical registers become typed SSA values, special registers become NVIDIA intrinsics, and memory-space qualifiers become LLVM address-space annotations.
| SSIR | LLVM IR |
| Registers (R0, R1, …) | Logical operand with type |
| Special registers (%tid, %ctaid) | NV-specific intrinsics |
| Memory spaces (global, shared, local) | Explicit address spaces |
| Arithmetic/logic operators | LLVM IR equivalents |
| Control flow (CondBra/Bra) | CondBr/Br instructions |
| Barriers (BAR.SYNC) | Sync intrinsics |
CuLifter maps SASS tensor operations to LLVM representations (Table 5). MMA instructions use NVIDIA’s standard llvm.nvvm.mma.* intrinsics; tensor memory operations lower to typed loads/stores with address-space annotations; tensor reductions decompose into scalar equivalents. We support SASS tensor types .F16, .F16x2, .BF16, .BF16x2, .TF32, .I8, and .I32, mapped to LLVM vector types , , , , and respectively.
| SASS operation | LLVM representation |
| MMA multiply-accumulate | @llvm.nvvm.mma.* |
| Tensor load | load <type>, addrspace(N) |
| Tensor store | store <type>, addrspace(N) |
| Tensor reduction (sum) | scalar FMA decomposition |
| Tensor reduction (max) | scalar compare/select |
5.6. Architecture Support
We support SM75 (Turing), SM90 (Hopper), and SM100/SM120 (Blackwell), with legacy support for SM35 (Kepler) and SM52 (Maxwell). Architecture support is localized to Stages 1 and 2: the parser handles different instruction encoding and predicate semantics, and the pattern library adds generation-specific idioms. Type analysis and LLVM generation are architecture independent. Adding a new generation requires extending the parser and pattern library only.
The main evolution across generations is the convergence mechanism. Pre-Volta architectures use reconvergence stacks (SSY/SYNC); Turing and later use convergence barriers (BSSY/BSYNC). We handle both by abstracting convergence as opaque synchronization intrinsics.
6. Evaluation
6.1. Methodology
We evaluate CuLifter on eight benchmark suites totaling 24,437 GPU functions (a function is either a __global__ or __device__ function) across 919 cubins (often a whole library archive, which may contains multiple kernels whose entry point is denoted as __global__ ).
CuLifter operates at the per function level, Table 6 reports the per-suite function counts and lifting results. Table 7 reports the kernel counts inherited from the original benchmark description, and the corresponding breakdown by application domain, spanning Linear Algebra, ML/DNN, Simulation, Science, Image/Vision, Graph, Crypto, and Infrastructure workloads. All results in this section are based on SM90 (Hopper).111Due to limited GPU resources, we focused evaluation on SM90. SM75 results are qualitatively similar. We verified correctness on SM100/SM120 (Blackwell) but omit those results for space.
Execution correctness. We evaluate type recovery against ground truth from debug information (DWARF) and source-level kernel signatures. A kernel passes execution correctness if its lifted IR and run on the same inputs, matches the original binary output (integer: bitwise; float: relative tolerance 10-5 for Float32, 10-2 for Float16).
Type accuracy. CuLifter achieves 100% kind accuracy (integer/float/bool classification) across all 24,437 functions. Exact accuracy (fully matching source type, including width) ranges from 77% to 87% depending on benchmark suite. The gap stems from register-level width mismatches introduced by the compiler—e.g., a 64-bit pointer loaded as two 32-bit halves—not from type confusion between different type families. Per-type-class accuracy is 100% for Bool, Float32, and Int32; lower values for Int64/Float64 are again due to splitting/merging artifacts. Coverage (fraction of ground-truth variables with an inferred type) exceeds 98% in all suites.
| Suite | Desc | Domain(s) | # Functions | # Lifted | Rate |
| HeCBench (Jin2023HeCBench) | open-source benchmark suite spanning 10 application domains (ML/DNN, Science, Simulation, etc.) | All 10 Domains | 773 | 773 | 100% |
| cuBLAS | closed-source BLAS library kernels from NVIDIA’s cuBLAS SDK | Linear Algebra | 7,260 | 7,260 | 100% |
| CUTLASS (cutlass) | NVIDIA’s open-source CUDA Templates for Linear Algebra Subroutines, including GEMM and convolution kernel | Linear Algebra, ML/DNN, Science, Infra., Other | 641 | 635 | 99.1% |
| FlashAttention (flashattention) | optimized attention kernels from the FlashAttention library representing state-of-the-art ML inference workloads | ML/DNN | 4,952 | 4,952 | 100% |
| cuDNN CNN (cudnn) | closed-source deep learning library. This sub-library contains kernels related to convolutional neural network | ML/DNN | 46 | 46 | 100% |
| cuDNN Ops | general tensor-operation kernels from cuDNN’s Ops sub-library | ML/DNN | 2,410 | 2,410 | 100% |
| cuDNN Adv | machine learning kernels from cuDNN’s Adv sub-library | ML/DNN | 4,336 | 4,336 | 100% |
| CUDA SDK (cudasdk) | CUDA kernels from CUDA SDK examples extracted via cuObjDump (cuda_binary_utilities) | ML/DNN, Sort/Search, Infra., Other | 19 | 19 | 100% |
| Total | 24,437 | 24,431 | 99.98% |
| Application Domain | Count |
| Linear Algebra | 1,089 |
| ML/DNN | 390 |
| Infrastructure | 54 |
| Science | 52 |
| Simulation | 29 |
| Other | 20 |
| Image/Vision | 18 |
| Sort/Search | 12 |
| Crypto | 6 |
| Graph | 6 |
| Stencil/PDE | 5 |
| Total | 1,681 |
6.2. Lift Results
CuLifter successfully lifts 24,431 of 24,437 GPU functions (99.98%) across eight benchmark suites. Table 6 shows per‑suite success rates ranging from 99.1% (CUTLASS) to 100% (all others).
Function-level failures: Only six functions fail to lift, all in CUTLASS. The root cause is the QGMMA instruction (SM90 tensor-core variant), which our parser currently emits as an opaque intrinsic.
Cubin-level results: At the granularity of compiled binary archives (cubins), CuLifter processes 872 of 919 cubins (94.9%). The remaining 47 cubins time out after 900 seconds due to the size of code. Importantly, every individual function inside those cubins still lifts successfully—the timeout only prevents the batch processing of the entire archive. No parser error or type‑recovery breakdown occurs.
6.3. Type Propagation Characterization
We validate the type-recovery claims of Section 3.1 by measuring six metrics across all benchmark suites.
Instruction classification. Figure 5 shows the fraction of instructions in each recovery role per domain. Type-seeding instructions dominate across all suites (93.0% combined, ranging from 77% in CUDA SDK to 96% in cuDNN Adv). Most values receive their type directly from a seed instruction without requiring propagation. The CUDA SDK shows a higher transparent fraction (22.8%), reflecting simpler hand-written kernels. Although seeding covers most instructions, the type-transparent gap is sufficient to break correctness without propagation, as the ablation confirms (Section 6.4).
Type composition. Figure 6 shows that Int32 is the dominant type across all domains, driven by addressing and fixed-point computation. Float32 contributes a substantial share in compute-intensive domains (linear algebra, ML, science) but is much smaller in integer-heavy domains such as sort/search and crypto. Int64 accounts for only a small fraction because global-memory addressing operations that require 64-bit pointers are far less frequent than arithmetic and control instructions. The 16-bit float family (, overall) and TF32 appear only where tensor-core operations are present (ML, linear algebra); within that family, HMMA.*.F32.BF16 opcodes select BF16 inputs while HMMA.*.F32 selects half, and HMMA.*.F32.TF32 selects TF32, all on the same 32-bit register width. Int128 does not appear in the final IR despite being part of the type lattice (Section 4.1). SASS has no native 128-bit ALU; LDG.128/STG.128 are wide memory transactions that move four 32-bit values simultaneously. The lifter decomposes all 128-bit operations into four scalar operations, so Int128 serves as a memory-width constraint during propagation but is never emitted.
Type resolution. Figure 7 breaks down value-resolution outcomes.Across all suites, 90.5% of values are resolved directly by seed instructions, while propagation resolves another 4.9%, filling the gap left by type-transparent instructions. Type conflicts constitute 3.2% of total values, and 1.4% remain unresolved after propagation, typically because no type evidence reaches them through the def-use chain. These unresolved values receive the Int32 fallback (Section 4.1).
Propagation behavior. Figure 8(a) shows that type evidence usually travels only 1–2 hops from the nearest seed instruction, with a small tail extending to 4 hops. Local evidence is typically sufficient, while a smaller set of cases requires multi-hop propagation through longer transparent chains.
Convergence Each fixpoint iteration propagates type constraints along def-use edges and checks for new conflicts; the analysis terminates when no type set changes. Figure 8(b) shows that convergence is fast in practice: the mean iteration count ranges from 2.2 (CUDA SDK) to 4.6 (cuDNN Ops), with a maximum of 6 and a combined mean of 3.5. Long propagation chains are uncommon, and the analysis converges reliably across all benchmark suites.
6.4. Ablation Study
We isolate the contribution of each recovery step by selectively disabling it and measuring the effect on lifting success. We restrict this experiment to HeCBench (229 cubins, one CUDA kernel and its host_cpu.cpp driver per cubin) due to the cost of lifting and executing larger suites; the 229 unit count in this table is the number of host drivers exercised, not the function granularity of Table 6.
Execution correctness is evaluated via the x86 backend, which emulates GPU execution for validation (Section 7.2). Due to the backend’s own semantic limitations (MUFU approximation fidelity, missing texture operations, SIMT emulation), the baseline achieves 73.8% pass rate. However, all ablations face the same x86 limitations, so any drop from the baseline isolates the effect of the disabled step. Table 8 and Figure 9 show the results.
| What is disabled | IR Gen | x86 Pass | Effect | |
| B0 | Nothing (full pipeline) | 229 | 169 (73.8%) | baseline |
| B1a | PredicateSEL | 228 | 167 (72.9%) | 0.9 pp |
| B1b | Pattern normalization | 229 | 169 (73.8%) | no change |
| B2 | Seed only (no propagation) | 8 | 0 (0.0%) | 73.8 pp |
| B3 | No type recovery (all int32) | 8 | 0 (0.0%) | 73.8 pp |
Type propagation gates correct execution. Disabling propagation (B2) or removing type recovery entirely (B3) drops the x86 pass rate from 73.8% to 0%. Every kernel still produce LLVM IR, the IR is functionally broken: registers whose types require propagation receive incorrect assignments, and the resulting IR fails downstream (compile error in 162 kernels, runtime crash in 45). Disabling predicate-to-select lowering (B1a) or pattern normalization (B1b) has negligible effect on the pass rate, confirming that type propagation is the single recovery step that determines whether the lifted IR is semantically faithful.
Resilience to reduced seed coverage. Figure 10 progressively removes seed instructions in descending order of frequency. The more seed instructions are removed, the more obvious the decline in the ration will be. This shows that propagation compensates for a moderate amount of missing seed information but still depends on a core set of high-frequency seed instructions.
6.5. Type Boundary Analysis
Type boundaries are program points where constraint propagation detects an empty intersection, indicating that the compiler has reinterpreted a register’s rely almost every domain. The Crypto domain is the sole outlier which contains no floating point patterns, as a different type without an explicit conversion. By analyzing the instructions at these boundaries, we find that most are not due to incorrect type recovery but rather intentional compiler optimizations. Table 9 summarizes the main categories: bit-level manipulation of floating-point values (IEEE 754 field extraction, float reconstruction), fast-math approximation chains (integer reciprocal via I2FMUFU.RCPIADD3F2I), and branch merges that join semantically distinct type paths.
Figure 11 categorizes the boundary patterns across all domains. The data shows that classifiable patterns account for more than half of the boundaries in almost every domain. The Crypto domain is the sole outlier which contains no floating point patterns, as its kernels rely almost entirely on integer arithmetic.
| Category | Example pattern |
| Branch merge | a PHI node joins a float path with an integer path (LOP3 bit-manipulation). |
| Fast math chains | I2F MUFU.RCP IADD3 F2I. Fast integer division. |
| IEEE 754 field ops | mantissa_extraction: LOP3.LUT Rd, Rsrc, 0x7fffff, RZ. Isolate mantissa field. |
| Float reconstruction | LOP3.LUT R53, R43, 0x7f800000, RZ, 0xfc. Forces Inf/NaN via exponent saturation |
| Analysis artifacts | IMAD.WIDE produces a 64-bit result, but downstream only uses the 32-bit low as float. |
| Unclassified | Uncategorized type conflicts. |
6.6. Discussion
Parser completeness. The six function-level lift failures (in CUTLASS) are caused by unimplemented SM90 instruction variants, primarily QGMMA. Adding support requires extending the parser and lowering rules but does not require changes to the type-propagation or control-flow algorithms. Each new GPU generation introduces new instructions, so parser maintenance is an ongoing cost.
Type recovery is sound but incomplete. Our constraint propagation detects conflicts and inserts bitcasts, but it cannot recover types for values that are never consumed by a type-seeding instruction. These values receive the Int32 fallback, which preserves bit patterns but may misrepresent intent. In practice, 1.4% of values remain unresolved (Section 6.3).
Intra-procedural analysis. Type propagation and control-flow recovery operate within a single kernel function. Device-side function calls (rare in SASS, where the compiler aggressively inlines) are lifted as opaque calls without cross-function type propagation.
Semantic fidelity. The CPU backend achieves a 73.8% end-to-end pass rate on HeCBench (compile, execute, and bitwise/tolerance-checked output match against the original binary). The remaining 26.2% are kernels that lift cleanly but fail downstream because hardware-specific operations (MUFU approximations, texture sampling, SM90 cluster features) have no exact CPU equivalent. The CPU backend is intended for analysis and testing, not performance.
7. Use Cases
7.1. Memory Access Pattern Analysis
We demonstrate analysis of closed-source kernels on a kernel with stride-based thread-variant memory accesses, a common pattern in ML and HPC code that causes poor coalescing. The SASS below, recovered from a compiled binary, shows the first violation: non-contiguous accesses where Thread 0 reads input[0], Thread 1 reads input[5], Thread 2 reads input[10], and so on, forcing 32 memory transactions per warp instead of 1–2:
The second case accesses every other element (Thread 0: input[0], Thread 1: input[4], …), requiring 16 transactions for 16 threads:
Both cases can be identified from lifted IR by tracing def-use chains from thread indices through address computations. This analysis requires typed IR: without type recovery, the lifter cannot emit valid LLVM IR at all (Section 6.4), and even partially incorrect types would misclassify integer index arithmetic as floating-point operations, breaking the address pattern.
7.2. Cross-Platform Execution on x86 and ARM
Because the lifted output is standard LLVM IR, we can retarget it to non-GPU backends including x86-64 and aarch64 (ARM). We compile lifted IR with clang and execute on commodity hardware, enabling functional testing without a GPU.
Execution model. The CPU runtime emulates GPU execution for correctness validation, not performance: one OS thread per GPU logical thread, with CTAs executed sequentially. Kernel parameters are passed through constant memory matching SASS’s c[0][ARG_BASE + offset] layout, so the lifted function has a void() signature with no ABI dependency. The base offset varies across GPU generations (e.g., 0x160 on SM75, 0x210 on SM90, 0x360 on SM100/SM120) because NVIDIA expands the constant-memory prefix with each generation to accommodate additional system state. CuLifter reads the correct offset from ELF metadata (EIATTR_PARAM_CBANK) rather than hardcoding it, so the same runtime supports all architectures.
| GPU Concept | CPU Mapping |
| Grid (gridDim CTAs) | Sequential CTA loop |
| CTA (blockDim threads) | std::thread pool (1 per GPU thread) |
| Warp (32 lanes) | 32 OS threads + mutex-based shuffle/vote |
| Shared memory (per-CTA) | Global array (zeroed per CTA) |
| Local memory (per-thread) | Thread-local storage (TLS) array |
| Constant memory (5 banks) | TLS array (copied per thread) |
| Registers (R0–R255) | SSA values (register-allocated by LLVM) |
| Predicates (P0–P7) | i1 values in LLVM IR |
GPU semantic decomposition. GPU-specific operations have no direct CPU equivalent; we decompose them to portable scalar operations as summarized in Table 11. All decompositions use standard LLVM instructions and C++17 synchronization primitives, with no platform-specific intrinsics or inline assembly.
| GPU Operation | CPU Decomposition | Notes |
| HMMA/WGMMA (tensor core) | Scalar FMA loops | Correct for lane 0 |
| SHFL (warp shuffle) | Mutex-based exchange | Per-warp shared buffer |
| VOTE/BALLOT (warp vote) | Mutex-based ballot | Per-warp ballot buffer |
| BAR (barrier) | Condition-variable wait | Counting barrier |
| ATOM (atomics) | Scalar load+op+store | Sequential CTA model |
| MUFU (special math) | libm functions | +3 ULP bias for RCP |
| MEMBAR/DEPBAR (fences) | NOP | Ordering via mutex |
| LDG/STG (global mem) | inttoptr + load/store | |
| LDS/STS (shared mem) | GEP into shared array |
Numerical fidelity. The hardest semantic gap is NVIDIA’s hardware reciprocal (MUFU.RCP), which returns an approximation biased by approximately ULPs relative to IEEE 754. The compiler-emitted integer-division sequence depends on this bias: our runtime must bias the IEEE reciprocal result by ULPs to match, or exact IEEE division produces wrong quotients.
ARM (aarch64) support. Supporting ARM required changing only the TLS model annotation and the LLVM target triple. The entire lifting pipeline (parser, transform passes, type analysis, and IR generation) is target-independent and produces identical LLVM IR regardless of the host CPU. We validated on an ARM Grace Hopper (GH200) system; all failures matched x86 failures with no ARM-specific issue. ARM’s relaxed memory model does not affect correctness because inter-thread communication in the runtime uses std::mutex, which provides full acquire/release semantics on both architectures.
The goal of cross-platform execution is analysis and validation, not performance parity. Texture/surface operations and SM90 cluster-level features are not yet supported.
8. Related Work
8.1. CPU Binary Lifters and Their Assumptions
CPU binary lifters, including McSema (mcsema), BAP (cpu_binary:2011:bap), RetDec (retdec), and others (cpu_binary:2000:uqbt; cpu_binary:2012:sp; cpu_binary:2017:ase; cpu_binary:2020:pldi; cpu_binary:2020:sp; cpu_binary:2008:vx32), are mature tools for x86, ARM, and RISC-V. As discussed in Section 3, the assumptions these tools rely on (branch-based control flow, separate register files, documented ISAs) do not hold for SASS.
8.2. Type Recovery: CPU vs. GPU
CPU binary type recovery is well-studied, with two main lines of work. The first is constraint-based inference: TIE (tie) introduced type inference with intersection and union types, Retypd (retypd) advanced this direction with recursive types, polymorphism, and subtyping, and BinSub (binsub2024) reformulates subtyping-based inference using algebraic subtyping, achieving a 63 runtime improvement over Retypd. The second line applies probabilistic and learning-based techniques: OSPREY (osprey) recovers variables and data structures probabilistically, DEBIN (debin) and DIRTY (dirty2022) predict types and variable names from stripped binaries using learning-based models, Manta (manta2024) combines hybrid-sensitive static analysis with type inference for bug detection, and TYGR (tygr) infers types for stripped binaries via graph neural networks. All of these systems operate in settings where ISA and ABI structure provide strong type evidence through separate register files, explicit branch instructions, and documented calling conventions.
On SASS, that evidence is much weaker because integer, floating-point, pointer, and packed values all share a single architectural register file. GPU type recovery therefore becomes a propagation problem with conflict detection rather than a classification problem. We must also handle intentional type punning, routinely emitted by the GPU compiler for fast reciprocal and exponent manipulation, through explicit bitcast insertion at conflict boundaries.
8.3. Tools for SASS or GPU Binary Analysis
NVIDIA’s cuObjDump and nvdisasm (cuda_binary_utilities) provide disassembly and CFG extraction. NVBit (nvbit) is a dynamic binary instrumentation framework; SASSI (sassi) is a compile-time instrumentation framework. Ari et al. (cgo:ari:2019) demonstrate assembler generation with a Hardware Abstraction Layer. Unlike CuLifter, none of these recover analyzable LLVM IR from SASS.
Jeanmougin et al. (jeansrc23) build CFGs for SASS using abstract interpretation to detect divergence. Shoushtary et al. (shoushtary2024) reverse-engineer 20 SASS control-flow instructions for CFG construction. Sen et al. (SenVS23) apply pattern matching to SASS for memory-access analysis. In contrast, CuLifter translates SASS to typed LLVM IR, enabling compiler-level analysis and retargeting to other architectures.
8.4. Running CUDA on Non-NVIDIA Hardware
Prior work translates CUDA at various abstraction levels (mcuda; FCUDA; DBLP:journals/corr/abs-2112-10034; han2024atdaes; SoftCUDA; Volt; ocelot; zluda; FlexGrip; HaunzhiWorkshop). Source-level tools (MCUDA (mcuda), Hipify (hipify)) convert CUDA source to CPU loops or HIP. COX (DBLP:journals/corr/abs-2112-10034) and CuPBoP/SoftCUDA/Volt (han2024atdaes; SoftCUDA; Volt) operate on NVVM IR. Ocelot (ocelot) lifts PTX to LLVM IR. ZLUDA (zluda) translates PTX at runtime. FlexGrip (FlexGrip) reverse-engineers SASS for integer-only FPGA execution. CASS (arxiv:25:CASS) ports CUDA to AMD at source and assembly level. Unlike all of these, CuLifter targets SASS (the final binary) and recovers typed LLVM IR.
9. Conclusion
We presented CuLifter, a framework that recovers analyzable LLVM IR from NVIDIA SASS binaries. Our key findings are:
-
(1)
Type recovery is the key challenge of GPU binary lifting. SASS’s unified register file erases all type information. We show that type recovery on GPUs is a constraint-propagation problem with conflict detection. Our ablation confirms this: disabling type recovery drops the x86 pass rate from 73.8% to 0%, a 73.8 percentage-point drop, while disabling other recovery steps has negligible effect.
-
(2)
Predicated-to-SSA conversion adapts known techniques. We apply -function-based select insertion from the IA-64 literature, extended for SIMT-specific constructs (predicated EXIT, dual-predicate branches, convergence barriers).
-
(3)
Practical lifting across GPU generations is feasible. CuLifter supports SM75 through SM120 and successfully lifts 24,431 of 24,437 GPU functions (99.98%) across eight benchmark suites spanning open-source applications, vendor libraries, and optimized ML runtimes, demonstrating that opaque GPU binaries can be brought back into the compiler-analysis stack.
The lifted LLVM IR enables downstream analysis (memory access pattern detection) and cross-platform execution on x86 and ARM CPUs.